open Pervasives

module S =
  Set.Make
    (struct
      type t = string
      let compare = compare
    end)

module M =
  Map.Make
    (struct
      type t = string
      let compare = compare
    end)

module M' =
  Map.Make
    (struct
      type t = string * string 
      let compare = compare
    end)

type maybe =
  | Just of float
  | Nothing

let rec errout l = (output_string stderr (String.concat " " l);output_string stderr "\n";flush stderr)

(* options *)
let ovwr = ref false
let depth = ref false
let itmin = ref 0
let itmax = ref 0

let counter = ref 0
let newvar n v =
  incr counter;
  String.concat "" [List.hd (Str.split (Str.regexp "[.]") v);".";string_of_int n;".";string_of_int !counter] 

let rec output_line f s =
  output_string f s;
  output_char f '\n'

(* add edge to initial and final vector *)
let rec addev x i v add =
  try
    let i' = M.find x v in
    M.add x (add i i') v
  with Not_found -> M.add x i v
(* add edge to map *)
let rec adde x y a i m add =
  try
    let yai' = M.find x m in
    try
      let ai' = M.find y yai' in
      try
	let i' = M.find a ai' in
	M.add x (M.add y (M.add a (add i i') ai') yai') m
      with Not_found -> M.add x (M.add y (M.add a i ai') yai') m
    with Not_found -> M.add x (M.add y (M.singleton a i) yai') m
  with Not_found -> M.add x (M.singleton y (M.singleton a i)) m
let rec adde' x y ai m add =
  M.fold (fun a i m -> adde x y a i m add) ai m
(* add edges to map *)
let rec addes x y ai m =
  try
    let yai' = M.find x m in
    M.add x (M.add y ai yai') m
  with Not_found -> M.add x (M.singleton y ai) m

(* merge edges *)
let rec mergee ai ai' add =
  let f a i ai =
    try
      let i' = M.find a ai in
      M.add a (add i i') ai
    with Not_found -> M.add a i ai in
  M.fold f ai ai'

(* reverse the graph *)
let rec reverse m =
  let rec add_from x yai n =
    M.fold (fun y ai n -> addes y x ai n) yai n in
  M.fold add_from m M.empty

(* collect nodes of a certain depth n. *)
let rec collectd ci m n =
  let rec collectd' s t n =
    if n <= 0 || S.is_empty t then t else
    let s' = S.union s t in
    let rec f x t' =
      try
        M.fold (fun y _ t' -> if S.mem y s' then t' else S.add y t') (M.find x m) t'
      with Not_found -> t' in
    collectd' s' (S.fold f t S.empty) (n-1) in
  collectd' S.empty (M.fold (fun x _ s -> S.add x s) ci S.empty) (n-1)

(* collect reachable state from ci. *)
let rec collectr ci m =
  let rec collectr' s t =
    if S.is_empty t then s else
    let s' = S.union s t in
    let rec f x t' =
      try
        M.fold (fun y _ t' -> if S.mem y s' then t' else S.add y t') (M.find x m) t'
      with Not_found -> t' in
    collectr' s' (S.fold f t S.empty) in
  collectr' S.empty (M.fold (fun x _ s -> S.add x s) ci S.empty) 

(* collect all nodes *)
let rec collectall ci m cv =
  let s = M.fold (fun x _ s -> S.add x s) ci S.empty in
  let s = M.fold (fun x yai s -> S.add x (M.fold (fun y _ s -> S.add y s) yai s)) m s in
  let s = M.fold (fun x _ s -> S.add x s) cv s in
  s    

let rec absx a x = String.concat "" [a;  "\\" ;x]

(* split and merge and elimination *)
let rec sme_se ci m cv s add mul munit =
  (* calculate ci *)
  let rec edge_ci x i ci' =
    (* x in S *)
    if S.mem x s then
      let ci' =
	try 
	  let j = M.find x cv in
	  addev "-\\!" (mul i j) ci' add
	with Not_found -> ci' in
      try
	let yai = M.find x m in
	M.fold (fun y ai ci' -> M.fold (fun a j ci' -> addev (absx a y) (mul i j) ci' add) ai ci') yai ci'
      with Not_found -> ci'
    (* x not in S *)
    else M.add x i ci' in
  let ci' = M.fold edge_ci ci M.empty in
  (* calculate m *)
  let rec edge_m x yai m' =
    if S.mem x s then
      let rec edge_m2 y ai m' =
	(* S -> S *)
	if S.mem y s then
	    let zbj = try M.find y m with Not_found -> M.empty in
	    let j = try Just (M.find y cv) with Not_found -> Nothing in 
	    let rec edge_m4 a _ m' =
	      let m' = match j with
		| Just j -> adde (absx a y) "-\\!" a j m' add
	        | Nothing -> m' in
	     M.fold (fun z bj m' -> M.fold (fun b j m' -> adde (absx a y) (absx b z) a j m' add) bj m') zbj m' in
	  M.fold edge_m4 ai m'
	(* S -> non S *)
	else M.fold (fun a _ m' -> adde (absx a y) y a munit m' add) ai m' in
      M.fold edge_m2 yai m' 
    else
      let rec edge_m3 y ai m' =
	(* non S -> S *)
	if S.mem y s then
	  try 
	    let zbj = M.find y m in
	    let rec edge_m5 a i m' =
	      let m' =
		try
		  let j = M.find y cv in
		  adde x "-\\!" a (mul i j) m' add
		with Not_found -> m' in
	      M.fold (fun z bj m' -> M.fold (fun b j m' -> adde x (absx b z) a (mul i j) m' add) bj m') zbj m' in
	    M.fold edge_m5 ai m' 
	  with Not_found -> m'
	(* non S -> non S *)
	else adde' x y ai m add in
      M.fold edge_m3 yai m' in
  let m' = M.fold edge_m m M.empty in
  (* calculate cv *)
  let cv' =
    if M.exists (fun x _ -> S.mem x s) cv then M.add "-\\!" munit (M.filter (fun x _ -> not (S.mem x s)) cv) else cv in
  (ci',m',cv')

(* apply FPE to the set of a certain depth *)
let rec sbn ci m cv iter count add mul munit=
  if count > iter then (ci,m,cv) else
  let s = collectd cv (reverse m) count in
  if S.is_empty s then
    sbn ci m cv iter (count+1) add mul munit
  else
    let (ci,m,cv) = sme_se ci m cv s add mul munit in
    sbn ci m cv iter (count+1) add mul munit

(* apply FPE to all *)
let rec sbn_all ci m cv add mul munit=
  let s = collectall ci m cv in
  sme_se ci m cv s add mul munit 
  
let rec input_line2 f =
  try
    input_line f
  with End_of_file -> "\n"
(* read the input file *)
let rec read_file f add ston =
  let rec read_file3 f cv =
      let s = input_line2 f in
      if s="\n" then cv else 
      let ss = Str.split (Str.regexp "[ \t]+") s in
      match ss with
      | a::b::_ -> read_file3 f (addev a (ston b) cv add) 
      | _ -> cv in
  let rec read_file2 f m =
      let s = input_line2 f in
      if s="\n" then (m,M.empty) else
      let ss = Str.split (Str.regexp "[ \t]+") s in
      match ss with
      | a::b::c::d::_ -> read_file2 f (adde a b c (ston d) m add)
      | _ -> let cv = read_file3 f M.empty in (m,cv) in
  let rec read_file1 f ci =
      let s = input_line2 f in
      if s="\n" then (ci,M.empty,M.empty) else
      let ss = Str.split (Str.regexp "[ \t]+") s in
      match ss with
      | a::b::_ -> read_file1 f (addev a (ston b) ci add)
      | _ -> let (m,cv) = read_file2 f M.empty in (ci,m,cv) in
  read_file1 f M.empty

(* output file *)
let rec output_file f ci m cv s ntos =
  M.iter (fun x i -> if S.mem x s||true then output_line f (String.concat " " [x;ntos i])) ci;
  output_char f '\n';
  M.iter (fun x yai -> if S.mem x s||true then M.iter (fun y ai -> if S.mem y s||true then M.iter (fun a i -> output_line f (String.concat " " [x;y;a;ntos i])) ai) yai) m;
  output_char f '\n';
  M.iter (fun x i -> if S.mem x s || true then output_line f (String.concat " " [x;ntos i])) cv
    
(* main *)
let rec main add mul munit ntos ston =
  let rec options i =
    if i >= Array.length Sys.argv then () else
    if Sys.argv.(i) = "-o" then (ovwr := true; options (i+1)) else
    if Sys.argv.(i) = "-d" && i+2<Array.length Sys.argv then
      (depth := true; itmin := max 1 (int_of_string Sys.argv.(i+1));
       itmax := int_of_string Sys.argv.(i+2); options (i+3)) in
  let fn = Sys.argv.(1) in
  let ifs = open_in fn in
  let (ci,m,cv) = read_file ifs add ston in
  options 3;
  let ((ci,m,cv),s) =
    if Str.string_match (Str.regexp "-\\(fpe\\|sb\\|fwdl\\|fl\\|bwdr\\|br\\)") Sys.argv.(2) 0 then
      let (ci,m,cv) =
	if !depth then sbn ci m cv !itmax !itmin add mul munit
	else sbn_all ci m cv add mul munit in
      ((ci,m,cv),"fpe")
    else if Str.string_match (Str.regexp "-\\(bpe\\|sf\\|fwdr\\|fr\\|bwdl\\|bl\\)") Sys.argv.(2) 0 then
      let (cv,m,ci) =
	if !depth then sbn cv (reverse m) ci !itmax !itmin add mul munit
        else sbn_all cv (reverse m) ci add mul munit in
      ((ci,reverse m, cv),"bpe")
    else raise (Invalid_argument Sys.argv.(2)) in
  let reachable = S.inter (collectr ci m) (collectr cv (reverse m)) in
  let fm =
    if !ovwr then fn
    else String.concat "" [List.hd (Str.split (Str.regexp "[.]") (List.hd (List.rev (Str.split (Str.regexp "[/]") fn))));"_";s;".out"] in
  let ofs = open_out fm in
  output_file ofs ci m cv reachable ntos;
  close_in ifs;
  close_out ofs

let _ = main max (+.) 0. (string_of_float) (float_of_string) 
