open Pervasives

type t

external ml_initialize: unit -> t = "ml_initialize"
external ml_open: unit -> t = "ml_open"

external ml_put_double: t -> float -> unit = "ml_put_double"
external ml_put_function: t -> string -> int -> unit = "ml_put_function"
external ml_put_symbol: t -> string -> unit = "ml_put_symbol"
external ml_end_packet: t -> unit = "ml_end_packet"

external ml_wait_arrival: t -> unit = "ml_wait_arrival"
external ml_get_string: t -> string = "ml_get_string"

external ml_close: t -> unit = "ml_close"
external ml_deinitialize: t -> unit = "ml_deinitialize"

let plus = ref "Plus"
let times = ref "Times"
let leq = ref "LessEqual"

let fwd = ref true

module S =
  Set.Make
    (struct
      type t = string
      let compare = compare
    end)

module M =
  Map.Make
    (struct
      type t = string
      let compare = compare
    end)


module Mi =
  Map.Make
    (struct
      type t = int
      let compare = compare
    end)

module Msi =
  Map.Make
    (struct
      type t = string * int
      let compare = compare
    end)

module Mil =
  Map.Make
    (struct
      type t = int list
      let compare = compare
    end)

(* power *)
let rec pow x y =
  let rec pow2 x y z =
    if y <= 0 then z else pow2 x (y-1) (x*z) in
  pow2 x y 1

(* nested for *)
let rec nested_for depth itmin itmax f =
  let rec nested_for_sub d f l =
    if d <= 0 then f l
    else
      for i=itmin to itmax do
	nested_for_sub (d-1) f (i::l)
      done
  in
  nested_for_sub depth f []

(* if number of arguments!=1 then send *)
let rec ml_put_function_ns lp f n =
  if n!=1 then ml_put_function lp f n else ()

module StrToInt = struct
  let empty = (M.empty,0)
  let rec add x (m,n) =
    if M.mem x m then (m,n)
    else (M.add x n m,n+1) 
  let rec add_and_return x (m,n) =
    try
      let i = M.find x m in (i,(m,n))
    with Not_found -> (n, (M.add x n m,n+1))
  let rec find x (m,n) = M.find x m
  let rec cardinal (m,n) = n
end

let rec input_line2 f =
  try
    input_line f
  with End_of_file -> "\n"

(* read initial transitions *)
let rec read_pre fn =
  let rec read_file1 init nodes =
      let s = input_line2 fn in
      if s="\n" then (init,nodes) else
      let ss = Str.split (Str.regexp "[ \t]+") s in
      match ss with
      | x::p::_ ->
	let (i,nodes) = StrToInt.add_and_return x nodes in
	read_file1 (Mi.add i (float_of_string p) init) nodes 
      | _ -> (init,nodes) in
  read_file1 Mi.empty StrToInt.empty

let rec read_file2 fn add trans nodes = try
  let s = input_line fn in
  let ss = Str.split (Str.regexp "[ \t]+") s in
  match ss with
  | x::a::p::xs ->
    let (x,nodes) = StrToInt.add_and_return x nodes in
    let (xs,nodes) = List.fold_left (fun (l,nodes) x -> let (i,nodes) = StrToInt.add_and_return x nodes in (i::l,nodes)) ([],nodes) xs in
    read_file2 fn add (add x (a,List.length xs) (List.rev xs) (float_of_string p) trans) nodes  
  | _ -> read_file2 fn add trans nodes 
  with End_of_file -> (trans,nodes)

(* read input file that is searched as column->row *)
let rec inputcr fn =
  let (init,nodes) = read_pre fn in
  let rec addcr x an xs p trans = 
    try
      let xsxp = Msi.find an trans in
      try
	let xp = Mil.find xs xsxp in
        try
  	  let p' = Mi.find x xp in
	  Msi.add an (Mil.add xs (Mi.add x (p+.p') xp) xsxp) trans
        with Not_found -> Msi.add an (Mil.add xs (Mi.add x p xp) xsxp) trans
      with Not_found -> Msi.add an (Mil.add xs (Mi.singleton x p) xsxp) trans
    with Not_found -> Msi.add an (Mil.singleton xs (Mi.singleton x p)) trans
  in
  let (trans,nodes) = read_file2 fn addcr Msi.empty nodes in
  (init,trans,StrToInt.cardinal nodes)

(* read input file that is searched as row->column *)
let rec inputrc fn =
  let (init,nodes) = read_pre fn in
  let rec addrc x an xs p trans = 
    try
      let anxsp = Mi.find x trans in
      try
	let xsp = Msi.find an anxsp in
	try
	  let p' = Mil.find xs xsp in
	  Mi.add x (Msi.add an (Mil.add xs (p+.p') xsp) anxsp) trans
	with Not_found -> Mi.add x (Msi.add an (Mil.add xs p xsp) anxsp) trans
      with Not_found -> Mi.add x (Msi.add an (Mil.singleton xs p) anxsp) trans
    with Not_found -> Mi.add x (Msi.singleton an (Mil.singleton xs p)) trans in
  let (trans,nodes) = read_file2 fn addrc Mi.empty nodes in
  (init,trans,StrToInt.cardinal nodes)

let rec varij x y = String.concat "" ["i";string_of_int x;"j";string_of_int y]

(* main *)
let _ =
  (* read arguments *)
  let rec options i =
    if i >= Array.length Sys.argv then () 
    else if Sys.argv.(i) = "-f" then (fwd:=true;options (i+1)) 
    else if Sys.argv.(i) = "-b" then (fwd:=false;options (i+1)) 
    else if Sys.argv.(i) = "-plus" && i+1<Array.length Sys.argv then
      (plus:=Sys.argv.(i+1);options (i+2)) 
    else if Sys.argv.(i) = "-times" && i+1<Array.length Sys.argv then
      (times:=Sys.argv.(i+1);options (i+2)) in
  options 3;
  let fn1 = open_in (Sys.argv.(1)) in
  let fn2 = open_in (Sys.argv.(2)) in

  (* read input file that is searched by column->row *)
  let (initcr,transcr,sizecr) =
    if !fwd then inputcr fn1 
    else inputcr fn2 in

  (* read input file that is searched by row->column *)
  let (initrc,transrc,sizerc) =
    if !fwd then inputrc fn2 
    else inputrc fn1 in

  close_in fn1;
  close_in fn2;

  (* initialize mathematica *)
  let env = ml_initialize () in
  let lp = ml_open () in

  (* send packet to mathematica *)
  ml_put_function lp "EvaluatePacket" 1;
  ml_put_function lp "Resolve" 2;
  ml_put_function lp "Exists" 2;
  (* list of variables *)
  ml_put_function lp "List" (sizecr * sizerc);
  for i=0 to sizecr-1 do
    for j=0 to sizerc-1 do
      ml_put_symbol lp (String.concat "" ["i";string_of_int i;"j";string_of_int j])
    done;
  done;

  (* list of constraints *)
  ml_put_function lp "And" 3;

  (* for initial states *)
  (if !fwd then
      (ml_put_function_ns lp "And" (Mi.cardinal initcr);
       let rec f x p =
	 ml_put_function lp !leq 2;
	 ml_put_double lp p;
	 ml_put_function_ns lp !plus (Mi.cardinal initrc);
	 let rec g y q =
	   ml_put_function_ns lp !times 2;
	   ml_put_double lp q;
	   ml_put_symbol lp (varij y x) in
	 Mi.iter g initrc
       in
       Mi.iter f initcr)
  else
    (ml_put_function_ns lp "And" sizecr;
     for y=0 to sizecr-1 do
       ml_put_function lp !leq 2;
       ml_put_function_ns lp !plus (Mi.cardinal initrc);
       let rec f x p =
	 ml_put_function lp !times 2;
	 ml_put_double lp p;
	 ml_put_symbol lp (varij x y) in
       Mi.iter f initrc;
       try
	 ml_put_double lp (Mi.find y initcr)
       with Not_found -> ml_put_function lp !plus 0
     done
    ));

  (* for transitions *)
  (if !fwd then(
      ml_put_function_ns lp "And" (sizerc * Msi.cardinal transcr);
      for y=0 to sizerc-1 do
	let rec f (a,n) xsxp =
	  ml_put_function_ns lp "And" (Mil.cardinal xsxp);
	  let rec h xs xp =	   
	    ml_put_function_ns lp !leq 2;
	    ml_put_function_ns lp !plus (Mi.cardinal xp);
	    let rec g x p =
	      ml_put_function lp !times 2;
	      ml_put_double lp p;
	      ml_put_symbol lp (varij y x) in
	    Mi.iter g xp;
	    try
	      let ysq = Msi.find (a,n) (Mi.find y transrc) in
	      ml_put_function_ns lp !plus (Mil.cardinal ysq);
	      let rec h ys q =
	        ml_put_function_ns lp !times (1+n);
	        ml_put_double lp q;
	        List.iter2 (fun y' x' -> ml_put_symbol lp (varij y' x')) ys xs in
	      Mil.iter h ysq
	    with Not_found -> ml_put_function lp !plus 0;
	  in
	  Mil.iter h xsxp
	in
	Msi.iter f transcr
      done)
   else( 
     ml_put_function_ns lp "And" (Mi.cardinal transrc);
     let rec f x anxsp =
       ml_put_function_ns lp "And" (Msi.cardinal anxsp);
       let rec g (a,n) xsp =
	 ml_put_function_ns lp "And" (pow sizecr n);
	 try
	   let ysyq = Msi.find (a,n) transcr in
	   let rec h ys =
	     ml_put_function lp !leq 2;
	     ml_put_function_ns lp !plus (Mil.cardinal xsp);
	     let rec e xs p =
	       ml_put_function_ns lp !times (1+n);
	       ml_put_double lp p;
	       List.iter2 (fun x' y' -> ml_put_symbol lp (varij x' y')) xs ys in
	     Mil.iter e xsp;
	     try
	       let yq = Mil.find ys ysyq in
	       ml_put_function_ns lp !plus (Mi.cardinal yq);
	       let rec d y q =
		 ml_put_function lp !times 2;
		 ml_put_double lp q;
		 ml_put_symbol lp (varij y x)
	       in
	       Mi.iter d yq
	     with Not_found -> ml_put_function lp !plus 0
	   in
	   nested_for n 0 (sizecr-1) h 
	 with Not_found ->
	   let rec h ys =
	     ml_put_function lp !leq 2;
	     ml_put_function_ns lp !plus (Mil.cardinal xsp);
	     let rec e xs p =
	       ml_put_function_ns lp !times (n+1);
	       ml_put_double lp p;
	       List.iter2 (fun x' y' -> ml_put_symbol lp (varij x' y')) xs ys in
	     Mil.iter e xsp;
	     ml_put_function lp !plus 0 in
	   nested_for n 0 (sizecr-1) h
       in
       Msi.iter g anxsp
     in
     Mi.iter f transrc
   )
  );

  (* all variables are non-negative *)
  ml_put_function_ns lp "And" (sizecr * sizerc);
  for i=0 to sizecr-1 do
    for j=0 to sizerc-1 do
      ml_put_function lp "GreaterEqual" 2;
      ml_put_symbol lp (varij i j);
      ml_put_function lp !plus 0;
    done;
  done;
  
  (* region of variables *)
  ml_put_symbol lp "Reals";
  ml_end_packet lp;

  (* receive packet from mathematica *)
  ml_wait_arrival lp;
  let s = ml_get_string lp in

  (* output the existence of simulation *)
  print_endline s;
  
  (* end mathematica *)
  ml_close lp;
  ml_deinitialize env
