#define DEBUG
#ifdef DEBUG
#include <stack>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <mathlink.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/fail.h>

#define MATHPATH "/Applications/Mathematica.app/Contents/MacOS/Mathematica"


#ifdef DEBUG
using namespace std;
stack<int> args;
void output_blanks(int n){
  for (int i=0; i<n; i++)printf("  ");
}
#endif

extern "C" value ml_initialize(value unit){
  MLEnvironment env;
  env = MLInitialize(NULL);
  if (env == NULL) caml_failwith("MLInitialize error");
  return (value)env;
}

extern "C" value ml_open(value unit){
  char *linkargv[]={(char *)"-linkname", (char *)MATHPATH};

  MLINK lp;
  lp = MLOpen(2, linkargv);
  if(lp == NULL) caml_failwith("MLOpen error");
  return (value)lp;
}

extern "C" value ml_put_integer(value lp, value v){
  MLPutInteger((MLINK)lp, Int_val(v));
  
#ifdef DEBUG
  output_blanks(args.size());
  printf("Integer %d\n", Int_val(v));
  if (!args.empty()) args.top()--;
  while (!args.empty() && args.top()<=0) args.pop();
#endif
  
  return Val_unit;
}

extern "C" value ml_put_double(value lp, value v){
  MLPutDouble((MLINK)lp, Double_val(v));
  
  #ifdef DEBUG
  output_blanks(args.size());
  printf("Double %lf\n", Double_val(v));
  if (!args.empty()) args.top()--;
  while (!args.empty() && args.top()<=0) args.pop();
  #endif

  return Val_unit;
}

extern "C" value ml_put_function(value lp, value f, value i){
  MLPutFunction((MLINK)lp, String_val(f), Int_val(i));
  
  #ifdef DEBUG
  output_blanks(args.size());
  printf("Function %s %d\n", String_val(f), Int_val(i));
  if (!args.empty()) args.top()--;
  args.push(Int_val(i));
  while (!args.empty() && args.top()<=0) args.pop();
  #endif

  return Val_unit;
}

extern "C" value ml_put_symbol(value lp, value s){
  MLPutSymbol((MLINK)lp, String_val(s));
  
  #ifdef DEBUG
  output_blanks(args.size());
  printf("Symbol %s\n", String_val(s));
  if (!args.empty()) args.top()--;
  while (!args.empty() && args.top()<=0) args.pop();
  #endif

  return Val_unit;
}

extern "C" value ml_end_packet(value lp){
  MLEndPacket((MLINK)lp);

  return Val_unit;
}

extern "C" value ml_wait_arrival(value lp){
  while (MLNextPacket((MLINK)lp) != RETURNPKT)  MLNewPacket((MLINK)lp);

  return Val_unit;
}

extern "C" value ml_get_string(value lp){
  const char *s;
  value ostr;
  MLGetString((MLINK)lp,&s);
  ostr = caml_copy_string(s);
  MLReleaseString((MLINK)lp,s);
  return (ostr);
}

extern "C" value ml_close(value lp){
  MLClose((MLINK)lp);

  return Val_unit;
}

extern "C" value ml_deinitialize(value env){
  MLDeinitialize((MLEnvironment)env);

  return Val_unit;
}
